/***************************************************************************
 *                                                                         *
 *    LIBDSK: General floppy and diskimage access library                  *
 *    Copyright (C) 2004  John Elliott <jce@seasip.demon.co.uk>            *
 *                                                                         *
 *    This library is free software; you can redistribute it and/or        *
 *    modify it under the terms of the GNU Library General Public          *
 *    License as published by the Free Software Foundation; either         *
 *    version 2 of the License, or (at your option) any later version.     *
 *                                                                         *
 *    This library is distributed in the hope that it will be useful,      *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *    Library General Public License for more details.                     *
 *                                                                         *
 *    You should have received a copy of the GNU Library General Public    *
 *    License along with this library; if not, write to the Free           *
 *    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,      *
 *    MA 02111-1307, USA                                                   *
 *                                                                         *
 ***************************************************************************/

/* DSKSCAN is meant to read a diskette 'blind' - something like the scan mode
 * of ANADISK. */

#include <stdio.h>
#include <stdlib.h>
#include "libdsk.h"
#include "utilopts.h"
#include "formname.h"
#include <errno.h>

#ifdef CPM
#define AV0 "DSKSCAN"
#else
#define AV0 argv[0]
#endif

static dsk_format_t format = -1;
static	char *intyp = NULL, *outtyp = NULL;
static	char *incomp = NULL, *outcomp = NULL;
static	int inside = -1, outside = -1;
static  int idstep =  0, odstep  =  0;
static  int retries = 1;

int do_copy(char *infile, char *outfile);
int scan_cyl(DSK_PDRIVER indr, DSK_GEOMETRY *dg, 
	     DSK_PDRIVER outdr, dsk_pcyl_t cyl, dsk_phead_t head);

static void report(const char *s)
{
        printf("%s\r", s);
        fflush(stdout);
}

static void report_end()
{
        printf("\r%-79.79s\r", "");
        fflush(stdout);
}


int help(int argc, char **argv)
{
	fprintf(stderr, "Syntax: \n"
                       "      %s {options} in-image { out-image }\n",
			AV0);
	fprintf(stderr,"\nOptions are:\n"
		       "-itype <type>   type of input disc image\n"
                       "-otype <type>   type of output disc image\n"
                       "-iside <side>   Force side 0 or side 1 of input\n"
                       "-oside <side>   Force side 0 or side 1 of output\n"
		       "-retry <count>  Set number of retries on error\n"
		       "-idstep         Double-step when reading\n"
		       "-odstep         Double-step when writing\n"
		       "-format         Force a specified format name\n");
	fprintf(stderr,"\nDefault in-image type is autodetect."
		               "\nDefault out-image type is DSK.\n\n");
		
	fprintf(stderr, "eg: %s /dev/fd0 myfile1.DSK\n"
                        "    %s -side 1 /dev/fd0 myfile2.DSK\n" 
                        "    %s -md3 /dev/fd0 md3boot.dsk\n" 
                        "    %s -otype floppy myfile.DSK /dev/fd0\n", 
			AV0, AV0, AV0, AV0);
	valid_formats();
	return 1;
}


int main(int argc, char **argv)
{

	if (find_arg("--version", argc, argv) > 0) return version(); 
	if (argc < 2) return help(argc, argv);
	if (find_arg("--help",    argc, argv) > 0) return help(argc, argv);

	ignore_arg("-type", 2, &argc, argv);
	ignore_arg("-side", 2, &argc, argv);
	ignore_arg("-dstep", 2, &argc, argv);
	ignore_arg("-comp", 2, &argc, argv);

	intyp     = check_type("-itype", &argc, argv); 
        outtyp    = check_type("-otype", &argc, argv);
	incomp    = check_type("-icomp", &argc, argv); 
        outcomp   = check_type("-ocomp", &argc, argv);
        inside    = check_forcehead("-iside", &argc, argv);
        outside   = check_forcehead("-oside", &argc, argv);
	retries   = check_retry("-retry", &argc, argv);
	if (present_arg("-idstep", &argc, argv)) idstep = 1;
	if (present_arg("-odstep", &argc, argv)) odstep = 1;
	if (!outtyp) outtyp = "dsk";
        format    = check_format("-format", &argc, argv);
	args_complete(&argc, argv);
	if (argc < 3) return do_copy(argv[1], NULL);
	return do_copy(argv[1], argv[2]);
}



static char *op = "Opening";


int do_copy(char *infile, char *outfile)
{
	DSK_PDRIVER indr = NULL, outdr = NULL;
	dsk_err_t e;
	dsk_pcyl_t cyl;
	dsk_phead_t head;
	char *cmt = NULL;
	DSK_GEOMETRY dg;
	dsk_pcyl_t maxcyl;
	dsk_phead_t maxhead = 2;

        dsk_reportfunc_set(report, report_end);

	        e = dsk_open (&indr,  infile,  intyp, incomp);
	if (!e) e = dsk_set_retry(indr, retries);
	if (!e && inside >= 0) e = dsk_set_option(indr, "HEAD", inside);
	if (!e && idstep) e = dsk_set_option(indr, "DOUBLESTEP", 1);

	if (outfile)
	{
		if (!e) e = dsk_creat(&outdr, outfile, outtyp, outcomp);
		if (!e && outside >= 0) e = dsk_set_option(outdr, "HEAD", outside);
		if (!e && odstep) e = dsk_set_option(outdr, "DOUBLESTEP", 1);
		if (!e) e = dsk_set_retry(outdr, retries); 
	}

/* Try to guess geometry. If this fails the user has to specify an approximate
 * geometry manually. We need to do this to get a feel for the number of 
 * cylinders to scan; dsk_psecid() will fail if (eg) it tries to select a
 * 1.4Mb geometry on a 360k drive */
        if (format == -1)
        {
                op = "Identifying disc";
                if (!e) e = dsk_getgeom(indr, &dg);
        }
        else if (!e) e = dg_stdformat(&dg, format, NULL, NULL);

        op = "Scanning disc";
	if (!e)
	{
		maxcyl = dg.dg_cylinders + 4;
// Head is being forced. Do only one side.
		if (inside >= 0) maxhead = 1;
		if (!outdr)
		{
			printf("<?xml version=\"1.0\" ?>\n");
			printf("<disc>\n");
		}
		dsk_get_comment(indr, &cmt);
		if (outdr) dsk_set_comment(outdr, cmt);
		else if (cmt)
		{
			printf("  <comment><![CDATA[%s]]></comment>\n", cmt);
		}
		for (cyl = 0; cyl < maxcyl; cyl++)
		{
			for (head = 0; head < maxhead; head++)
			{
				if (!outdr) printf("  <track cylinder=\"%d\" head=\"%d\">\n", cyl, head);
				e = scan_cyl(indr, &dg, outdr, cyl, head);
				if (!outdr) printf("  </track>\n");
				if (e) goto abort;
			}
		}
abort:
		if (!outdr) printf("</disc>\n");
	}
	if (indr)  dsk_close(&indr);
	if (outdr) dsk_close(&outdr);
	if (e)
	{
		fprintf(stderr, "\n%s: %s\n", op, dsk_strerror(e));
		return 1;
	}
	return 0;
}


/* Number of times a specified sector can be seen before we are certain that the
 * disc has made a complete rotation and nothing has been missed. */
#define PASSES 2

int scan_cyl(DSK_PDRIVER indr, DSK_GEOMETRY *dg, 
	     DSK_PDRIVER outdr, dsk_pcyl_t cyl, dsk_phead_t head)
{
	unsigned char have_sector[256];
	dsk_err_t err = DSK_ERR_OK;
	DSK_FORMAT sector_id;
	DSK_FORMAT fmt_track[256];
	DSK_FORMAT *pfmt = fmt_track;
	dsk_psect_t sector_count, osec;
	unsigned secnum;
	unsigned buflen = 128;

	memset(have_sector, 0, sizeof(have_sector));

	dg->dg_noskip = 1;
	// Guess data rate and recording mode
	for (dg->dg_datarate = RATE_HD; dg->dg_datarate <= RATE_ED; ++dg->dg_datarate)
	{
		for (dg->dg_fm = 0; dg->dg_fm < 2; ++dg->dg_fm)
		{
			err = dsk_psecid(indr, dg, cyl, head, &sector_id);
			if (!err) break;	
		}	
		if (!err) break;
	}
	if (err)
	{
		if (!outdr) printf("    <!-- Found nothing -->\n");
		return DSK_ERR_OK;
	}
	if (!outdr)
	{
		switch(dg->dg_datarate)
		{
			case RATE_HD: printf("    <datarate>500</datarate>\n"); break;
			case RATE_DD: printf("    <datarate>250</datarate>\n"); break;
			case RATE_SD: printf("    <datarate>300</datarate>\n"); break;
			case RATE_ED: printf("    <datarate>1000</datarate>\n"); break;
		}
		if (dg->dg_fm)	printf("    <encoding>fm</encoding>\n");
		else		printf("    <encoding>mfm</encoding>\n");
	}
/* Now scour the track and print any sector headers we find */
	sector_count = 0;
	do	
	{
		++have_sector[sector_id.fmt_sector];
		if (have_sector[sector_id.fmt_sector] == 1)
		{
			if (!outdr)
			{
				printf("    <sector>\n");
				printf("      <cyl>%d</cyl>",   sector_id.fmt_cylinder);
				if (sector_id.fmt_cylinder != cyl)
					printf(" <!-- Does not match location on disc -->");
				printf("\n");
				printf("      <hd>%d</hd>",     sector_id.fmt_head);
				if (sector_id.fmt_head != head)
					printf(" <!-- Does not match location on disc -->");
				printf("\n");
				printf("      <sec>%d</sec>\n", sector_id.fmt_sector);
				printf("      <size>%d</size>\n", sector_id.fmt_secsize);
				printf("    </sector>\n");
			}
			memcpy(pfmt, &sector_id, sizeof(DSK_FORMAT));
			++pfmt;
			++sector_count;
			if (sector_id.fmt_secsize > buflen) buflen = sector_id.fmt_secsize;
		}
		err = dsk_psecid(indr, dg, cyl, head, &sector_id);
		if (err) return err;
	}
	while (have_sector[sector_id.fmt_sector] < 1 + PASSES);
	if (outdr && sector_count)
	{
		unsigned char *buf = dsk_malloc(buflen);
		if (!buf) return DSK_ERR_NOMEM;

/* Format a track with the correct number and size of sectors */
		osec = dg->dg_sectors;
		dg->dg_sectors = sector_count;
		op = "Formatting";
		err = dsk_pformat(outdr, dg, cyl, head, fmt_track, 0xE5);
		dg->dg_sectors = osec;
		if (!err) for (secnum = 0; secnum < sector_count; secnum++)
		{
			int deleted = 0;
			op = "Reading";
			printf("dsk_xread %d %d  %d,%d,%d,%d sectors=%d\n",
				cyl, head,
				fmt_track[secnum].fmt_cylinder,
				fmt_track[secnum].fmt_head,
				fmt_track[secnum].fmt_sector,
				fmt_track[secnum].fmt_secsize, sector_count);

			err = dsk_xread(indr, dg, buf, cyl, head,
					fmt_track[secnum].fmt_cylinder,
					fmt_track[secnum].fmt_head,
					fmt_track[secnum].fmt_sector,
					fmt_track[secnum].fmt_secsize, &deleted);
			// XXX Need to be able to write sectors with errors,
			// at least to CPCEMU DSK 
			if (err == DSK_ERR_DATAERR) err = DSK_ERR_OK;
			if (err) break;
			op = "Writing";
			err = dsk_xwrite(outdr, dg, buf, cyl, head,
					fmt_track[secnum].fmt_cylinder,
					fmt_track[secnum].fmt_head,
					fmt_track[secnum].fmt_sector,
					fmt_track[secnum].fmt_secsize, deleted);
			if (err) break;	
		}
		dsk_free(buf);
		if (err) return err;
	}
	return DSK_ERR_OK;
}
